# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
# import os
# import sys
# sys.path.insert(0, os.path.abspath('.'))

import hpmoc

# -- Project information -----------------------------------------------------

project = 'HEALPix Multi-Order Coordinate Library (HPMOC)'
copyright = '2021, Stefan Trklja Countryman'
author = 'Stefan Trklja Countryman'

# The short X.Y version
version = ".".join(hpmoc.__version__.split(".")[:2])
# The full version, including alpha/beta/rc tags
release = hpmoc.__version__


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'nbsphinx',  # include jupyter notebooks
    'sphinx.ext.autodoc',
    'sphinx.ext.doctest',
    'sphinx.ext.todo',
    'sphinx.ext.coverage',
    'sphinx.ext.graphviz',
    # 'sphinx.ext.mathjax', # excluded in LLAMA, likely performance issues?
    'sphinx.ext.imgmath',
    # 'sphinx.ext.viewcode', # do not want to include source in docs
    #'sphinx.ext.githubpages',  # add .nojekyll and CNAME
    'sphinx.ext.napoleon',  # understand NumPy and Google docstring syntax
    #'sphinxarg.ext',  # understand CLIs https://sphinx-argparse.readthedocs.io; conda install sphinx-argparse
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
# source_suffix = ['.rst', '.md']
source_suffix = '.rst'

# The master toctree document.
master_doc = 'index'

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = None

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = []

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = None


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'sphinx_rtd_theme'

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#
# html_theme_options = {}
html_theme_options = {
    'canonical_url': 'http://hpmoc.stc.sh',
    # 'analytics_id': 'UA-XXXXXXX-1',  #  Provided by Google in your dashboard
    # 'logo_only': False,
    'display_version': True,
    'prev_next_buttons_location': 'bottom',
    'style_external_links': True,
    # 'style_nav_header_background': 'black',
    # Toc options
    'collapse_navigation': False,
    'sticky_navigation': True,
    'navigation_depth': 5,
    # 'includehidden': True,
    # 'titles_only': False
}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

# -- Options for LaTeX output ------------------------------------------------

latex_show_urls = 'footnote'
latex_engine = 'xelatex'

# preamble example taken from
# https://sphinxguide.readthedocs.io/en/latest/sphinx_basics/settings.html
latex_elements = {
    'printindex': '\\footnotesize\\raggedright\\printindex',
    'preamble': r'''
        \usepackage[utf8]{inputenc}
        \usepackage{amssymb}
        \usepackage{newunicodechar}

        \newunicodechar{ℝ}{$\mathbb{R}$}
        \newunicodechar{ρ}{$\rho$}
        \newunicodechar{Δ}{$\Delta$}
        \newunicodechar{δ}{$\delta$}
        \newunicodechar{Ω}{$\Omega$}
        \newunicodechar{ω}{$\omega$}
        \newunicodechar{Θ}{$\Theta$}
        \newunicodechar{θ}{$\theta$}
        \newunicodechar{ϵ}{$\epsilon$}
        \newunicodechar{ν}{$\nu$}
        \newunicodechar{ₙ}{$_{n}$}
        \newunicodechar{ₕ}{$_{h}$}
        \newunicodechar{ᶜ}{$^{c}$}
        \newunicodechar{ˢ}{$^{s}$}
        \newunicodechar{ᵉ}{$^{e}$}
        \newunicodechar{ᵐ}{$^{m}$}
        \newunicodechar{ᵖ}{$^{p}$}
    ''',

    # ADD UNICODE SUPPORT
    # 'utf8extra': r'''
    #     \ifdefined\DeclareUnicodeCharacter
    #         \ifdefined\DeclareUnicodeCharacterAsOptional
    #         \else
    #             \DeclareUnicodeCharacter{00A0}{\nobreakspace}\
    #             \DeclareUnicodeCharacter{3C1}{\nobreakspace}\
    #         \fi
    #     \fi
    # '''
    # The paper size ('letterpaper' or 'a4paper').
    #
    # 'papersize': 'letterpaper',

    # The font size ('10pt', '11pt' or '12pt').
    #
    # 'pointsize': '10pt',

    # Additional stuff for the LaTeX preamble.
    #
    # 'preamble': '',

    # Latex figure (float) alignment
    #
    # 'figure_align': 'htbp',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    (master_doc, 'llama.tex', 'Low-Latency Algorithm for Multi-messenger Astrophysics (LLAMA)',
     'Stefan Trklja Countryman', 'manual'),
]

# -- Extension configuration -------------------------------------------------

# add graphviz directives to each filehandler class found in llama
graphviz_output_format = 'svg'  # use SVG images for webpage output

